#ifdef	HAVE_UNISTD_H
#include	<unistd.h>
#else
#include	<stdio.h>
#endif
#ifdef	STDC_HEADERS
#include	<stdlib.h>
#endif
#include	<fcntl.h>
#include	<time.h>
#include	<sys/stat.h>

#include	"tailor.h"
#include	"gzip.h"
#include	"lzw.h"

#ifndef	uch
#define	uch	unsigned char
#endif

#define	isalpha(c)	((uch)((c)-'A')<=(uch)('Z'-'A')||(uch)((c)-'a')<=(uch)('z'-'a'))
#define	isupper(c)	((uch)((c)-'A')<=(uch)('Z'-'A'))

char libgzip_version[] = "gzip-1.2.4" LIBGZIP_VERSION;


char *progname = "gzip_thread";       /* program name */
int to_stdout = 0;    /* output to stdout (-c) */
int force = 1;        /* don't ask questions, compress links (-f) */
int no_name = -1;     /* don't save or restore the original file name */
int no_time = -1;     /* don't save or restore the original file time */
int list = 0;         /* list the file contents (-l) */
int maxbits = BITS;   /* max bits per code for LZW */
int method = DEFLATED;/* compression method */
int level = 6;        /* compression level */
int verbose = 0;      /* be verbose (-v) */
int quiet = 1;        /* be very quiet (-q) */
int exit_code = OK;   /* program exit code */
int save_orig_name;   /* set if original name must be saved */
int last_member;      /* set for .zip and .Z files */
int test = 0;         /* test .gz file integrity */
long time_stamp;      /* original time stamp (modification time) */
#ifdef	HUMAN68K
int	backup = 0;       /* force backup mode */
int shortname = 0;    /* shortname extention */
#endif

int  ifd = _THREAD_BIAS_;                  /* input file descriptor */
int  ofd = _THREAD_BIAS_;                  /* output file descriptor */
unsigned outcnt;           /* bytes in output buffer */

DECLARE(uch, inbuf,  INBUFSIZ +INBUF_EXTRA);
DECLARE(uch, outbuf, OUTBUFSIZ+OUTBUF_EXTRA);
DECLARE(ush, d_buf,  DIST_BUFSIZE);
DECLARE(uch, window, 2L*WSIZE);
#ifndef MAXSEG_64K
    DECLARE(ush, tab_prefix, 1L<<BITS);
#else
    DECLARE(ush, tab_prefix0, 1L<<(BITS-1));
    DECLARE(ush, tab_prefix1, 1L<<(BITS-1));
#endif

unsigned insize;           /* valid bytes in inbuf */
unsigned inptr;            /* index of next byte to be processed in inbuf */
long bytes_in;             /* number of input bytes */
long bytes_out;            /* number of output bytes */
char ifname[MAX_PATH_LEN]; /* input file name */
char ofname[MAX_PATH_LEN]; /* output file name */
int  remove_ofname = 0;    /* remove output file on error */

int (*work) OF((int infile, int outfile)) = zip; /* function to call */

static int	part_nb = 0;

/* ========================================================================
 * Check the magic number of the input file and update ofname if an
 * original name was given and to_stdout is not set.
 * Return the compression method, -1 for error, -2 for warning.
 * Set inptr to the offset of the next byte to be processed.
 * Updates time_stamp if there is one and --no-time is not used.
 * This function may be called repeatedly for an input file consisting
 * of several contiguous gzip'ed members.
 * IN assertions: there is at least one remaining compressed member.
 *   If the member is a zip file, it must be the only one.
 */
static int get_method(in)
    int in;        /* input file descriptor */
{
    uch flags;     /* compression flags */
    char magic[2]; /* magic header */
    ulg stamp;     /* time stamp */

    ifd = in;

    /* If --force and --stdout, zcat == cat, so do not complain about
     * premature end of file: use try_byte instead of get_byte.
     */
    if (force && to_stdout) {
	magic[0] = (char)try_byte();
	magic[1] = (char)try_byte();
	/* If try_byte returned EOF, magic[1] == 0xff */
    } else {
	magic[0] = (char)get_byte();
	magic[1] = (char)get_byte();
    }
    method = -1;                 /* unknown yet */
    part_nb++;                   /* number of parts in gzip file */
    header_bytes = 0;
    last_member = RECORD_IO;
    /* assume multiple members in gzip file except for record oriented I/O */

    if (memcmp(magic, GZIP_MAGIC, 2) == 0
        || memcmp(magic, OLD_GZIP_MAGIC, 2) == 0) {

	method = (int)get_byte();
	if (method != DEFLATED) {
	    fprintf(stderr,
		    "%s: %s: unknown method %d -- get newer version of gzip\n",
		    progname, ifname, method);
	    exit_code = ERROR;
	    return -1;
	}
	work = unzip;
	flags  = (uch)get_byte();

	if ((flags & ENCRYPTED) != 0) {
	    fprintf(stderr,
		    "%s: %s is encrypted -- get newer version of gzip\n",
		    progname, ifname);
	    exit_code = ERROR;
	    return -1;
	}
	if ((flags & CONTINUATION) != 0) {
	    fprintf(stderr,
	   "%s: %s is a a multi-part gzip file -- get newer version of gzip\n",
		    progname, ifname);
	    exit_code = ERROR;
	    if (force <= 1) return -1;
	}
	if ((flags & RESERVED) != 0) {
	    fprintf(stderr,
		    "%s: %s has flags 0x%x -- get newer version of gzip\n",
		    progname, ifname, flags);
	    exit_code = ERROR;
	    if (force <= 1) return -1;
	}
	stamp  = (ulg)get_byte();
	stamp |= ((ulg)get_byte()) << 8;
	stamp |= ((ulg)get_byte()) << 16;
	stamp |= ((ulg)get_byte()) << 24;
	if (stamp != 0 && !no_time) time_stamp = stamp;

	(void)get_byte();  /* Ignore extra flags for the moment */
	(void)get_byte();  /* Ignore OS type for the moment */

	if ((flags & CONTINUATION) != 0) {
	    unsigned part = (unsigned)get_byte();
	    part |= ((unsigned)get_byte())<<8;
	    if (verbose) {
		fprintf(stderr,"%s: %s: part number %u\n",
			progname, ifname, part);
	    }
	}
	if ((flags & EXTRA_FIELD) != 0) {
	    unsigned len = (unsigned)get_byte();
	    len |= ((unsigned)get_byte())<<8;
	    if (verbose) {
		fprintf(stderr,"%s: %s: extra field of %u bytes ignored\n",
			progname, ifname, len);
	    }
	    while (len--) (void)get_byte();
	}

	/* Get original file name if it was truncated */
	if ((flags & ORIG_NAME) != 0) {
	    if (no_name || (to_stdout && !list) || part_nb > 1) {
		/* Discard the old name */
		char c; /* dummy used for NeXTstep 3.0 cc optimizer bug */
		do {c=get_byte();} while (c != 0);
	    } else {
		/* Copy the base name. Keep a directory prefix intact. */
                char *p = basename(ofname);
                char *base = p;
		for (;;) {
		    *p = (char)get_char();
		    if (*p++ == '\0') break;
		    if (p >= ofname+sizeof(ofname)) {
			error("corrupted input -- file name too large");
		    }
		}
                /* If necessary, adapt the name to local OS conventions: */
                if (!list) {
                   MAKE_LEGAL_NAME(base);
		   if (base) list=0; /* avoid warning about unused variable */
                }
	    } /* no_name || to_stdout */
	} /* ORIG_NAME */

	/* Discard file comment if any */
	if ((flags & COMMENT) != 0) {
	    while (get_char() != 0) /* null */ ;
	}
	if (part_nb == 1) {
	    header_bytes = inptr + 2*sizeof(long); /* include crc and size */
	}

    } else if (memcmp(magic, PKZIP_MAGIC, 2) == 0 && inptr == 2
	    && memcmp((char*)inbuf, PKZIP_MAGIC, 4) == 0) {
	/* To simplify the code, we support a zip file when alone only.
         * We are thus guaranteed that the entire local header fits in inbuf.
         */
        inptr = 0;
	work = unzip;
	if (check_zipfile(in) != OK) return -1;
	/* check_zipfile may get ofname from the local header */
	last_member = 1;

    } else if (memcmp(magic, PACK_MAGIC, 2) == 0) {
	work = unpack;
	method = PACKED;

    } else if (memcmp(magic, LZW_MAGIC, 2) == 0) {
	work = unlzw;
	method = COMPRESSED;
	last_member = 1;

    } else if (memcmp(magic, LZH_MAGIC, 2) == 0) {
	work = unlzh;
	method = LZHED;
	last_member = 1;

    } else if (force && to_stdout && !list) { /* pass input unchanged */
	method = STORED;
	work = copy;
        inptr = 0;
	last_member = 1;
    }
    if (method >= 0) return method;

    if (part_nb == 1) {
	fprintf(stderr, "\n%s: %s: not in gzip format\n", progname, ifname);
	exit_code = ERROR;
	return -1;
    } else {
	WARN((stderr, "\n%s: %s: decompression OK, trailing garbage ignored\n",
	      progname, ifname));
	return -2;
    }
}


static int open_and_check(char* file, int mode)
{
	int	fd;

	if ((fd = open(file, mode)) != -1) {
		method = get_method(fd);
		if (method < 0) {
			close(fd);
			return -1;
		}
	}
	return fd;
}

int	_gzip_open(char* file, int mode)
{
	int	fd;
	int	len = strlen(file);

	/* o̓t@C */
	if (len >= sizeof ifname) return -1;
	strcpy(ifname, file);
	strcpy(ofname, "pseudo-pipe");
    clear_bufs(); /* clear input and output buffers */
    part_nb = 0;

	/* ܂A^ꂽt@CŃI[vĂ݂ */
	if ((fd = open_and_check(ifname, mode)) != -1) return fd;

	/* I[vłȂꍇ͊gq⊮ */
	if (file[len-1] != 'z' && file[len-1] != 'Z') {
		/* ܂ ".gz" tĂ݂ */
		strcat(ifname, ".gz");
		if ((fd = open_and_check(ifname, mode)) != -1) return fd;
		/*  ".Z" tĂ݂ */
		strcat(strcpy(ifname, file), ".Z");
		if ((fd = open_and_check(ifname, mode)) != -1) return fd;

		/* Ō̂Q "GZ" ɕύXĂ݂ */
		strcpy(strcpy(ifname, file) + len - 2, "GZ");
		if ((fd = open_and_check(ifname, mode)) != -1) return fd;
		/* Ō̂P "Z" ɕύXĂ݂ */
		strcpy(strcpy(ifname, file) + len - 1, "Z");
		if ((fd = open_and_check(ifname, mode)) != -1) return fd;
	}
	strcpy(ifname, file);
	return -1;
}


static int	backup_open(char* file, int mode, int attr)
{
	struct stat	st;

	if (backup && stat(file, &st) == 0) {
		char newfile[PATH_MAX];

		strcat(strcpy(newfile, file), "~");
		chmod(file, st.st_mode | (S_IRUSR|S_IWUSR));
		if (rename(file, newfile) == 0) {
			chmod(newfile, st.st_mode);
		}
	}
	return open(file, mode, attr);
}


int	_gzip_create(char* file, int mode)
{
	int	fd;
	int	len = strlen(file);
	extern unsigned char* mbsrchr(unsigned char*, int);

	time_stamp = time(0);
	remove_ofname = 1;

	/* o̓t@C */
	if (len >= sizeof ifname) return -1;
	save_orig_name = 1;
	strcpy(ifname, file);
	strcpy(ofname, file);

	/*  "gz" łȂꍇ͊gq⊮s */
	if (stricmp(file + len - 2, "gz") != 0) {
		char* p = ofname;
		char* q;

		/* hCuXLbv */
		if (isalpha(p[0]) && p[1] == ':') p += 2;

		/* fBNgXLbv */
		while ((q = (char*)mbsrchr(p, '/')) || (q = (char*)mbsrchr(p, '\\'))) p = ++q;

		/* gq邩H */
		q = (char*)mbsrchr(p, '.');

		if (q && shortname) {
			/* gq shortname w莞 */
			if (q[1]) strcpy(q+2, (isupper(q[1])  ? "GZ" : "gz"));
			else      strcpy(q+1, (isupper(q[-1]) ? "GZ" : "gz"));
		}
		else {
			/* ȊO͒P ".gz" t */
			strcat(ofname, (isupper(ofname[len-1]) ? ".GZ" : ".gz"));
		}
		if ((fd = backup_open(ofname, mode, S_IREAD|S_IWRITE)) != -1) return fd;

		/* ŃI[vłȂꍇ͍Ō̂Q "gz" ɕύX */
		strcpy(strcpy(ofname, file) + len - 2,
		       (isupper(ofname[len-3]) ? "GZ" : "gz"));
		if ((fd = backup_open(ofname, mode, S_IREAD|S_IWRITE)) != -1) return fd;
	}
	else if (stricmp(file + len - 3, ".gz") == 0) {
		/* gq .gz ̎͒Pɐ؂l߂ */
		ifname[len - 3] = '\0';
	}
	else if (stricmp(file + len - 4, ".tgz") == 0) {
		/* gq .tgz ̎ .tar ɕϊ */
		strcpy(ifname + len - 3, (isupper(ifname[len-3]) ? "TAR" : "tar"));
	}
	else {
		save_orig_name = 0;
		strcpy(ifname, "pseudo-pipe");
	}

	/* ^ꂽt@Ĉ܂܎g */
	strcpy(ofname, file);
	return backup_open(ofname, mode, S_IREAD|S_IWRITE);
}


/* ========================================================================
 * Signal and error handler.
 */
RETSIGTYPE abort_gzip(void)
{
   if (remove_ofname) {
       thread_close(ofd);
       unlink (ofname);
   }
   exit(-1);
}


